/*
   RainbowCrack - a general propose implementation of Philippe Oechslin's
   faster time-memory trade-off technique.

   Copyright (C) Zhu Shuanglei <shuanglei@hotmail.com>
*/

#ifdef _WIN32
	#pragma warning(disable : 4786)
#endif

#include "GChainWalkContext.h"

#include <ctype.h>
#ifdef _WIN32
	#pragma comment(lib, "libeay32.lib")
#endif


GCChainWalkContext::GCChainWalkContext(unsigned int nChainCounter,
				       unsigned int nRainbowChainCount,
				       unsigned int nRainbowChainLen)
{
  m_nChainCounter = nChainCounter;
  m_nRainbowChainCount = nRainbowChainCount;
  m_nRainbowChainLen = nRainbowChainLen;

#ifdef _WIN32
  cmutex = CreateMutex(NULL, FALSE, NULL);
  if (cmutex == NULL)
    perror("CreateMutex failed\n");
#else
  pthread_mutex_init(&cmutex, NULL);
#endif

  m_nStartClock = clock();
}

GCChainWalkContext::~GCChainWalkContext()
{
#ifdef _WIN32
  CloseHandle(cmutex);
#else
  pthread_mutex_destroy(&cmutex);
#endif

  PrintStats();
}

bool GCChainWalkContext::LoadCharset(string sName)
{
	if (sName == "byte")
	{
		int i;
		for (i = 0x00; i <= 0xff; i++)
			m_PlainCharset[i] = i;
		m_nPlainCharsetLen = 256;
		m_sPlainCharsetName = sName;
		m_sPlainCharsetContent = "0x00, 0x01, ... 0xff";
		return true;
	}

	vector<string> vLine;
	if (ReadLinesFromFile("charset.txt", vLine))
	{
		unsigned int i;
		for (i = 0; i < vLine.size(); i++)
		{
			// Filter comment
			if (vLine[i][0] == '#')
				continue;

			vector<string> vPart;
			if (SeperateString(vLine[i], "=", vPart))
			{
				// sCharsetName
				string sCharsetName = TrimString(vPart[0]);
				if (sCharsetName == "")
					continue;

				// sCharsetName charset check
				bool fCharsetNameCheckPass = true;
				unsigned int j;
				for (j = 0; j < sCharsetName.size(); j++)
				{
					if (   !isalpha(sCharsetName[j])
						&& !isdigit(sCharsetName[j])
						&& (sCharsetName[j] != '-'))
					{
						fCharsetNameCheckPass = false;
						break;
					}
				}
				if (!fCharsetNameCheckPass)
				{
					fprintf(stderr, "invalid charset name %s in charset "
						"configuration file\n", sCharsetName.c_str());
					continue;
				}

				// sCharsetContent
				string sCharsetContent = TrimString(vPart[1]);
				if (sCharsetContent == "" || sCharsetContent == "[]")
					continue;
				if (sCharsetContent[0] != '[' || sCharsetContent[sCharsetContent.size() - 1] != ']')
				{
					fprintf(stderr, "invalid charset content %s in charset "
						"configuration file\n", sCharsetContent.c_str());
					continue;
				}
				sCharsetContent = sCharsetContent.substr(1, sCharsetContent.size() - 2);
				if (sCharsetContent.size() > 256)
				{
					fprintf(stderr, "charset content %s too long\n",
						sCharsetContent.c_str());
					continue;
				}

				//printf("%s = [%s]\n", sCharsetName.c_str(), sCharsetContent.c_str());

				// Is it the wanted charset?
				if (sCharsetName == sName)
				{
					m_nPlainCharsetLen = sCharsetContent.size();
					memcpy(m_PlainCharset, sCharsetContent.c_str(), m_nPlainCharsetLen);
					m_sPlainCharsetName = sCharsetName;
					m_sPlainCharsetContent = sCharsetContent;
					return true;
				}
			}
		}
		fprintf(stderr, "charset %s not found in charset.txt\n",
			sName.c_str());
	}
	else
		perror("can't open charset configuration file\n");

	return false;
}

//////////////////////////////////////////////////////////////////////

bool GCChainWalkContext::SetHashRoutine(string sHashRoutineName)
{
	CHashRoutine hr;
	hr.GetHashRoutine(sHashRoutineName, m_pHashRoutine, m_nHashLen);
	if (m_pHashRoutine != NULL)
	{
		m_sHashRoutineName = sHashRoutineName;
		return true;
	}
	else
		return false;
}

HASHROUTINE GCChainWalkContext::GetHashRoutine() {
  return m_pHashRoutine;
}

bool GCChainWalkContext::SetPlainCharset(string sCharsetName, int nPlainLenMin, int nPlainLenMax)
{
	// m_PlainCharset, m_nPlainCharsetLen, m_sPlainCharsetName, m_sPlainCharsetContent
	if (!LoadCharset(sCharsetName))
		return false;

	// m_nPlainLenMin, m_nPlainLenMax
	if (nPlainLenMin < 1 || nPlainLenMax > MAX_PLAIN_LEN || nPlainLenMin > nPlainLenMax)
	{
		fprintf(stderr, "invalid plaintext length range: %d - %d\n",
			nPlainLenMin, nPlainLenMax);
		return false;
	}
	m_nPlainLenMin = nPlainLenMin;
	m_nPlainLenMax = nPlainLenMax;

	// m_nPlainSpaceUpToX
	m_nPlainSpaceUpToX[0] = 0;
	uint64 nTemp = 1;
	unsigned int i;
	for (i = 1; i <= m_nPlainLenMax; i++)
	{
		nTemp *= m_nPlainCharsetLen;
		if (i < m_nPlainLenMin)
			m_nPlainSpaceUpToX[i] = 0;
		else
			m_nPlainSpaceUpToX[i] = m_nPlainSpaceUpToX[i - 1] + nTemp;
	}

	// m_nPlainSpaceTotal
	m_nPlainSpaceTotal = m_nPlainSpaceUpToX[m_nPlainLenMax];

	return true;
}

bool GCChainWalkContext::SetRainbowTableIndex(int nRainbowTableIndex)
{
	if (nRainbowTableIndex < 0)
		return false;
	m_nRainbowTableIndex = nRainbowTableIndex;
	m_nReduceOffset = 65536 * nRainbowTableIndex;

	return true;
}

string GCChainWalkContext::GetHashRoutineName()
{
	return m_sHashRoutineName;
}

unsigned int GCChainWalkContext::GetHashLen()
{
	return m_nHashLen;
}

unsigned char *GCChainWalkContext::GetPlainCharset()
{
	return m_PlainCharset;
}

unsigned int GCChainWalkContext::GetPlainCharsetLen()
{
	return m_nPlainCharsetLen;
}

unsigned int GCChainWalkContext::GetPlainLenMin()
{
	return m_nPlainLenMin;
}

unsigned int GCChainWalkContext::GetPlainLenMax()
{
	return m_nPlainLenMax;
}

string GCChainWalkContext::GetPlainCharsetName()
{
	return m_sPlainCharsetName;
}

string GCChainWalkContext::GetPlainCharsetContent()
{
	return m_sPlainCharsetContent;
}

uint64 *GCChainWalkContext::GetPlainSpaceUpToX()
{
	return m_nPlainSpaceUpToX;
}

uint64 GCChainWalkContext::GetPlainSpaceTotal()
{
	return m_nPlainSpaceTotal;
}

int GCChainWalkContext::GetRainbowTableIndex()
{
	return m_nRainbowTableIndex;
}

uint64 GCChainWalkContext::GetReduceOffset()
{
	return m_nReduceOffset;
}

unsigned int GCChainWalkContext::GetRainbowChainLen()
{
	return m_nRainbowChainLen;
}

void GCChainWalkContext::Dump()
{
	printf("hash routine: %s\n", m_sHashRoutineName.c_str());
	printf("hash length: %d\n", m_nHashLen);

	printf("plain charset: ");
	unsigned int i;
	for (i = 0; i < m_nPlainCharsetLen; i++)
	{
		if (isprint(m_PlainCharset[i]))
			printf("%c", m_PlainCharset[i]);
		else
			printf("?");
	}
	printf("\n");

	printf("plain charset in hex: ");
	for (i = 0; i < m_nPlainCharsetLen; i++)
		printf("%02x ", m_PlainCharset[i]);
	printf("\n");

	printf("plain length range: %d - %d\n", m_nPlainLenMin, m_nPlainLenMax);
	printf("plain charset name: %s\n", m_sPlainCharsetName.c_str());
	//printf("plain charset content: %s\n", m_sPlainCharsetContent.c_str());
	//for (i = 0; i <= m_nPlainLenMax; i++)
	//	printf("plain space up to %d: %s\n", i, uint64tostr(m_nPlainSpaceUpToX[i]).c_str());
	printf("plain space total: %s\n", uint64tostr(m_nPlainSpaceTotal).c_str());

	printf("rainbow table index: %d\n", m_nRainbowTableIndex);
	printf("reduce offset: %s\n", uint64tostr(m_nReduceOffset).c_str());
	printf("\n");
}

void GCChainWalkContext::SetFile(FILE *f) {
  m_file = f;
}

FILE *GCChainWalkContext::GetFile() {
  return m_file;
}

void GCChainWalkContext::PrintStats() const {
  clock_t current = clock();
  int nSecond = (current - m_nStartClock) / CLOCKS_PER_SEC;

  printf("%d of %d rainbow chains generated (%d m %d s)\n",
	 m_nChainCounter,
	 m_nRainbowChainCount,
	 nSecond / 60,
	 nSecond % 60);
}

bool GCChainWalkContext::IncCounterAndStop(unsigned int val=1) {
  m_nChainCounter += val;

  if ((m_nChainCounter % 100000) < val)
    PrintStats();

  return m_nChainCounter >= m_nRainbowChainCount - 1;
}

void GCChainWalkContext::AcquireLock() {
#ifdef _WIN32
  DWORD res = WaitForSingleObject(cmutex, INFINITE);
  if (res != WAIT_OBJECT_0)
    perror("Cannot acquire mutex\n");
#else
  pthread_mutex_lock(&cmutex);
#endif
}

void GCChainWalkContext::ReleaseLock() {
#ifdef _WIN32
  ReleaseMutex(cmutex);
#else
  pthread_mutex_unlock(&cmutex);
#endif
}
