/*
   RainbowCrack - a general propose implementation of Philippe Oechslin's
   faster time-memory trade-off technique.

   Copyright (C) Zhu Shuanglei <shuanglei@hotmail.com>
*/

#ifdef _WIN32
	#pragma warning(disable : 4786)
#endif

#include "LChainWalkContext.h"

#include <ctype.h>
#include <openssl/rand.h>
#ifdef _WIN32
	#pragma comment(lib, "libeay32.lib")
#endif


LCChainWalkContext::LCChainWalkContext(GCChainWalkContext *gctx)
{
  m_gctx = gctx;
}

LCChainWalkContext::~LCChainWalkContext()
{
}

void LCChainWalkContext::GenerateRandomIndex()
{
	RAND_bytes((unsigned char*)&m_nIndex, 8);
	m_nIndex = m_nIndex % m_gctx->m_nPlainSpaceTotal;
}

void LCChainWalkContext::SetIndex(uint64 nIndex)
{
	m_nIndex = nIndex;
}

void LCChainWalkContext::SetHash(unsigned char* pHash)
{
	memcpy(m_Hash, pHash, m_gctx->m_nHashLen);
}

void LCChainWalkContext::IndexToPlain()
{
	int i;
	for (i = m_gctx->m_nPlainLenMax - 1;
	     i >= (int) m_gctx->m_nPlainLenMin - 1; i--)
	{
		if (m_nIndex >= m_gctx->m_nPlainSpaceUpToX[i])
		{
			m_nPlainLen = i + 1;
			break;
		}
	}

	uint64 nIndexOfX = m_nIndex - m_gctx->m_nPlainSpaceUpToX[m_nPlainLen - 1];

	/*
	// Slow version
	for (i = m_nPlainLen - 1; i >= 0; i--)
	{
		m_Plain[i] = m_PlainCharset[nIndexOfX % m_nPlainCharsetLen];
		nIndexOfX /= m_nPlainCharsetLen;
	}
	*/

	// Fast version
	for (i = m_nPlainLen - 1; i >= 0; i--)
	{
#ifdef _WIN32
		if (nIndexOfX < 0x100000000I64)
			break;
#else
		if (nIndexOfX < 0x100000000llu)
			break;
#endif

		m_Plain[i] = m_gctx->m_PlainCharset[nIndexOfX % m_gctx->m_nPlainCharsetLen];
		nIndexOfX /= m_gctx->m_nPlainCharsetLen;
	}
	unsigned int nIndexOfX32 = (unsigned int)nIndexOfX;
	for (; i >= 0; i--)
	{
		//m_Plain[i] = m_PlainCharset[nIndexOfX32 % m_nPlainCharsetLen];
		//nIndexOfX32 /= m_nPlainCharsetLen;

		unsigned int nPlainCharsetLen = m_gctx->m_nPlainCharsetLen;
		unsigned int nTemp;
#ifdef _WIN32
		__asm
		{
			mov eax, nIndexOfX32
			xor edx, edx
			div nPlainCharsetLen
			mov nIndexOfX32, eax
			mov nTemp, edx
		}
#else
		__asm__ __volatile__ (	"mov %2, %%eax;"
					"xor %%edx, %%edx;"
					"divl %3;"
					"mov %%eax, %0;"
					"mov %%edx, %1;"
					: "=m"(nIndexOfX32), "=m"(nTemp)
					: "m"(nIndexOfX32), "m"(nPlainCharsetLen)
					: "%eax", "%edx"
					);
#endif
		m_Plain[i] = m_gctx->m_PlainCharset[nTemp];
	}
}

void LCChainWalkContext::PlainToHash()
{
	m_gctx->m_pHashRoutine(m_Plain, m_nPlainLen, m_Hash);
}

void LCChainWalkContext::HashToIndex(int nPos)
{
	m_nIndex = (*(uint64*)m_Hash + m_gctx->m_nReduceOffset + nPos) % m_gctx->m_nPlainSpaceTotal;
}

uint64 LCChainWalkContext::GetIndex()
{
	return m_nIndex;
}

string LCChainWalkContext::GetPlain()
{
	string sRet;
	int i;
	for (i = 0; i < m_nPlainLen; i++)
	{
		char c = m_Plain[i];
		if (c >= 32 && c <= 126)
			sRet += c;
		else
			sRet += '?';
	}

	return sRet;
}

string LCChainWalkContext::GetBinary()
{
	return HexToStr(m_Plain, m_nPlainLen);
}

string LCChainWalkContext::GetPlainBinary()
{
	string sRet;
	sRet += GetPlain();
	unsigned int i;
	for (i = 0; i < m_gctx->m_nPlainLenMax - m_nPlainLen; i++)
		sRet += ' ';

	sRet += "|";

	sRet += GetBinary();
	for (i = 0; i < m_gctx->m_nPlainLenMax - m_nPlainLen; i++)
		sRet += "  ";

	return sRet;
}

string LCChainWalkContext::GetHash()
{
	return HexToStr(m_Hash, m_gctx->m_nHashLen);
}

bool LCChainWalkContext::CheckHash(unsigned char* pHash)
{
	if (memcmp(m_Hash, pHash, m_gctx->m_nHashLen) == 0)
		return true;

	return false;
}
