/*
   RainbowCrack - a general propose implementation of Philippe Oechslin's
   faster time-memory trade-off technique.

   Copyright (C) Zhu Shuanglei <shuanglei@hotmail.com>
*/

#ifdef _WIN32
        #define _CRT_SECURE_NO_WARNINGS 1
	#pragma warning(disable : 4786)
#endif

#ifdef _WIN32
	#include <windows.h>
	#include <io.h>
	#include <fcntl.h>
#else
	#include <unistd.h>
        #include <cassert>
        #include <pthread.h>
#endif
#include "GChainWalkContext.h"
#include "LChainWalkContext.h"


void Usage() {
  Logo();

  printf("usage: rtgen-mt hash_algorithm \\\n");
  printf("                plain_charset plain_len_min plain_len_max \\\n");
  printf("                rainbow_table_index \\\n");
  printf("                rainbow_chain_length rainbow_chain_count \\\n");
  printf("                file_title_suffix\\\n");
  printf("                number_of_threads\n");
  printf("\n");

  CHashRoutine hr;
  printf("hash_algorithm:       available: %s\n", hr.GetAllHashRoutineName().c_str());
  printf("plain_charset:        use any charset name in charset.txt here\n");
  printf("                      use \"byte\" to specify all 256 characters as the charset of the plaintext\n");
  printf("plain_len_min:        min length of the plaintext\n");
  printf("plain_len_max:        max length of the plaintext\n");
  printf("rainbow_table_index:  index of the rainbow table\n");
  printf("rainbow_chain_length: length of the rainbow chain\n");
  printf("rainbow_chain_count:  count of the rainbow chain to generate\n");
  printf("file_title_suffix:    the string appended to the file title\n");
  printf("                      add your comment of the generated rainbow table here\n");
  printf("number_of_threads:    number of worker threads to create\n");

  printf("\n");
  printf("example: rtgen-mt lm alpha 1 7 0 100 16 test 2\n");
  printf("         rtgen-mt md5 byte 4 4 0 100 16 test 2\n");
  printf("         rtgen-mt sha1 numeric 1 10 0 100 16 test 4\n");
}


#ifdef _WIN32
DWORD WINAPI KernelFunc(LPVOID arg) {
#else
void *KernelFunc(void *arg) {
#endif
  GCChainWalkContext *gctx = (GCChainWalkContext *) arg;
  LCChainWalkContext lctx(gctx);
  const unsigned int num_buffer_chains = 128;
  uint64 chain[2 * num_buffer_chains];
  bool stop = false;
  unsigned int lcount = 0;

  while (1) {
    lctx.GenerateRandomIndex();
    chain[lcount++] = lctx.GetIndex();

    unsigned int nPos;
    for (nPos = 0;
	 nPos < gctx->GetRainbowChainLen() - 1;
	 nPos++) {
      lctx.IndexToPlain();
      lctx.PlainToHash();
      lctx.HashToIndex(nPos);
    }

   chain[lcount++] = lctx.GetIndex();

   if (lcount == 2 * num_buffer_chains) {
     gctx->AcquireLock();

     if (fwrite(chain, 8, 2 * num_buffer_chains,
		gctx->GetFile()) != 2 * num_buffer_chains) {
       perror("disk write fail\n");
       break;
     }

     if (gctx->IncCounterAndStop(num_buffer_chains))
       stop = true;

     gctx->ReleaseLock();

     if (stop)
       break;

     lcount = 0;
   }
  }

#ifdef _WIN32
  return 0;
#else
  pthread_exit(NULL);
#endif
}


int main(int argc, char* argv[]) {
  if (argc != 10) {
    Usage();
    return 0;
  }

  string sHashRoutineName  = argv[1];
  string sCharsetName      = argv[2];
  int nPlainLenMin         = atoi(argv[3]);
  int nPlainLenMax         = atoi(argv[4]);
  int nRainbowTableIndex   = atoi(argv[5]);

  unsigned int nRainbowChainLen     = atoi(argv[6]);
  unsigned int nRainbowChainCount   = atoi(argv[7]);
  string sFileTitleSuffix  = argv[8];
  unsigned int num_threads = atoi(argv[9]);

  if (!(num_threads < MAX_NUM_THREADS)) {
    printf("Number of threads limited to %d",
	   MAX_NUM_THREADS);
    return 1;
  }

  // nRainbowChainCount check
  if (nRainbowChainCount >= 134217728) {
    printf("this will generate a table larger than 2GB, which "
	   "is not supported\n");
    printf("please use a smaller rainbow_chain_count(less than "
	   "134217728)\n");
    return 1;
  }

  // FileName
  char szFileName[256];
  sprintf(szFileName, "%s_%s#%d-%d_%d_%dx%d_%s.rt",
	  sHashRoutineName.c_str(),
	  sCharsetName.c_str(),
	  nPlainLenMin,
	  nPlainLenMax,
	  nRainbowTableIndex,
	  nRainbowChainLen,
	  nRainbowChainCount,
	  sFileTitleSuffix.c_str());

  // Open file
  fclose(fopen(szFileName, "a"));
  FILE* file = fopen(szFileName, "r+b");
  if (file == NULL) {
    printf("failed to create %s\n", szFileName);
    return 1;
  }

  // Check existing chains
  unsigned int nDataLen = GetFileLen(file);
  nDataLen = nDataLen / 16 * 16;
  if (nDataLen == nRainbowChainCount * 16) {
    printf("precomputation of this rainbow table already "
	   "finished\n");
    fclose(file);
    return 1;
  }
  if (nDataLen > 0)
    printf("continuing from interrupted precomputation...\n");
  fseek(file, nDataLen, SEEK_SET);

  // Global context
  GCChainWalkContext gctx(nDataLen / 16,
			  nRainbowChainCount,
			  nRainbowChainLen);

  gctx.SetFile(file);

  // Setup GCChainWalkContext
  if (!gctx.SetHashRoutine(sHashRoutineName)) {
    printf("hash routine %s not supported\n",
	   sHashRoutineName.c_str());
    fclose(file);
    return 1;
  }
  if (!gctx.SetPlainCharset(sCharsetName, nPlainLenMin,
			    nPlainLenMax)) {
    fclose(file);
    return 1;
  }
  if (!gctx.SetRainbowTableIndex(nRainbowTableIndex)) {
    printf("invalid rainbow table index %d\n",
	   nRainbowTableIndex);
    fclose(file);
    return 1;
  }

  gctx.Dump();

#ifdef _WIN32
  DWORD dwThreadIdArray[MAX_NUM_THREADS];
  HANDLE threads[MAX_NUM_THREADS];
#else
  pthread_t threads[MAX_NUM_THREADS];
#endif

  // Generate rainbow table
  printf("generating (from %u threads)...\n", num_threads);

  unsigned int i;

  // Create num_threads threads
  for(i = 0; i < num_threads; ++i) {
#ifdef _WIN32
    threads[i] = CreateThread(NULL, 0, KernelFunc,
			      &gctx, 0, &dwThreadIdArray[i]);
    if (threads[i] == NULL) {
      perror("create_thread() failed\n");
#else
    int rc = pthread_create(&threads[i], NULL,
			    KernelFunc, (void *) &gctx);
    if (rc) {
      fprintf(stderr, "pthread_create() failed, %d\n", rc);
#endif
      return 1;
    }
  }

  // Wait for threads completions
#ifdef _WIN32
  WaitForMultipleObjects(num_threads, threads, TRUE, INFINITE);
  for (i = 0; i < num_threads; ++i)
    CloseHandle(threads[i]);
#else
  for (i = 0; i < num_threads; ++i) {
    pthread_join(threads[i], NULL);
  }
#endif

  // Close file
  fclose(file);

  // Truncate file
#ifdef _WIN32
  int fd = _open(szFileName, _O_RDWR);
  if (fd == -1) {
    perror("Open file failed");
    return 1;
  }
  if (_chsize(fd, nRainbowChainCount * 16) == -1)
    perror("Cannot truncate file");
  _close(fd);
#else
  if (truncate(szFileName, nRainbowChainCount * 16) != 0)
    perror("Cannot truncate file");
#endif

  return 0;
}
